local input = {}

-- can be exposed to user configuration
-- each action is a key with the inputs
-- that trigger it as elements in the table associated
local keyInputs = {
  up = {"w","up", "l"},
  left = {"a", "left", ","},
  right = {"d", "right", "."},
  shoot = {"space", " ", "return"}
}

-- returns true if any of the keys are down / pressed
-- (for a particular action with multiple associated keys)
function input.down(keys)
  for _, key in ipairs(keys) do
    if love.keyboard.isDown(key) then
      return true
    end
  end
  return false
end

-- returns a table to track user inputs
-- would be called on change to keyInputs
function input.load()
  local inputs = {}
  for action, _ in pairs(keyInputs) do
    inputs[action] = {}
    inputs[action].down = false
    inputs[action].released = false
    inputs[action].pressed = false
  end
  return inputs
end

-- updates the inputs table for this action
function input.listen(inputs, action)
  local down = input.down(keyInputs[action])
  -- mark this action as pressed if it was not previously
  -- pressed should trigger on the first time step
  -- a key for an action is pressed down, then be
  -- consumed, and then stay false till all keys for
  -- the action are released, at which point it may
  -- be triggered again
  if (not inputs[action].pressed) and (not inputs[action].down) and down then
    inputs[action].pressed = true
  end
  -- similarly in reverse for released
  if (not inputs[action].released) and inputs[action].down and (not down) then
    inputs[action].released = true
  end
  -- setting released/pressed to false should happen upon consumption of
  -- the inputs table, or on discard after proccessing a game tick
  -- as it is possible for multiple ticks of this function before
  -- a game tick responds to them
  inputs[action].down = down
end

-- collects all user inputs into the inputs table
function input.collect(inputs)
  for action, _ in pairs(keyInputs) do
    input.listen(inputs, action)
  end
end

-- discards all unclaimed pressed/released inputs
-- called after a game tick is finished
function input.discard(inputs)
  for action, _ in pairs(keyInputs) do
    inputs[action].released = false
    inputs[action].pressed = false
  end
end

return input
