local array = require 'lib.L-C.array'
local list = require 'lib.L-C.list'

-- responsible for creating individual physics based entities like a ship
-- or bullets, but not for managing them or creating levels
local factory = {}

local BULLET_LIFESPAN = 100

function factory.ship(gameState, x, y, size, callback)
  local world = gameState.physics.world
  -- create a triangle
  local shipBody = world:newPolygonCollider(
    {-size,-(2*size/3), 0,size, size,-(2*size/3)}
  )
  shipBody:setPosition(x, y)
  local weaponWidth = size*0.2
  -- place each weapon on each side of the ship
  shipWeapon1 = world:newRectangleCollider(
    x - size/2 - weaponWidth/2,
    -size/3,
    weaponWidth,
    size*1.2)
  shipWeapon2 = world:newRectangleCollider(
    x + size/2 - weaponWidth/2,
    -size/3,
    weaponWidth,
    size*1.2)
  -- place joint in centre of player, the weapons and the main
  -- body should collide with each other
  world:addJoint('WeldJoint', shipWeapon1, shipBody, x, y, true)
  world:addJoint('WeldJoint', shipWeapon2, shipBody, x, y, true)
  -- make collisions inellastic
  shipBody:setRestitution(0.2)
  shipWeapon1:setRestitution(0.2)
  shipWeapon2:setRestitution(0.2)
  -- minimise unwanted spinning, especially on welded weapons
  shipBody:setAngularDamping(20)
  shipWeapon1:setAngularDamping(20)
  shipWeapon2:setAngularDamping(20)
  -- make movement slow down
  shipBody:setLinearDamping(0.1)
  shipWeapon1:setLinearDamping(0.1)
  shipWeapon2:setLinearDamping(0.1)
  shipBody.shoot = function(shipBody)
    local x, y = shipBody:getPosition()
    -- rotate by 90 degrees to get angle player is shooting at
    local angle = shipBody:getAngle() + math.pi/2
    -- calculate the velocity to give the bullets based on this angle
    local sx, sy = 100*math.cos(angle), 100*math.sin(angle)
    -- place the bullets on each side of the ship's weapons
    for _, v in ipairs({ -math.pi/8, math.pi/8 }) do
      local dx = size*1 * math.cos(angle + v)
      local dy = size*1 * math.sin(angle + v)
      local bullet = world:newCircleCollider(x + dx, y + dy, size*0.1)
      -- apply the impulse to the bullets
      bullet:applyLinearImpulse(sx, sy)
      bullet:setRestitution(0.01)
      bullet:setAngularDamping(0.01)
      bullet:setLinearDamping(0.01)
      -- ensure collisions are not lost due to the high speed
      bullet:setBullet(true)
      bullet.lifespan = BULLET_LIFESPAN
      -- give the bullet an update function TODO use metatables
      bullet.update = function(bullet)
        local dx, dy = bullet:getLinearVelocity()
        if math.abs(dx) < 2 and math.abs(dy) < 2 then
          bullet.lifespan = bullet.lifespan - 1
        end
        if bullet.lifespan < 0 then
          -- fire all callbacks that listen to removal
          gameState.physics.callbacks.remove(gameState, bullet, "bullet")
          -- destroy this bullet in the box2d world
          bullet:destroy()
          -- return true to stop tracking this entity
          return true
        end
        return false
      end
      gameState.physics.callbacks.new(gameState, bullet, "bullet")
    end
  end
  shipBody.update = function(shipBody)
    -- don't delete this ship ever TODO use metatables
    return false
  end
  gameState.physics.callbacks.new(gameState, shipBody, "player")
  return shipBody
end

return factory
