local wf = require 'lib.windfield'
local factory = require 'logic.world.factory.factory'
local array = require 'lib.L-C.array'
local list = require 'lib.L-C.list'

-- responsible for managing the physics world,
-- not creating parts of it directly
local world = {}

local WORLD_TICK_TIMESTEP = 1/60

function world.load(gameState)
  -- create a blank world
  gameState.physics = {}
  -- create entities list to track things that need to be updated
  gameState.entities = list.new(array.new({}))
  gameState.physics.world = wf.newWorld(0, 0, true)
  -- setup a table for other logic to listen to the
  -- creation and deletion of physics objects
  gameState.physics.listeners = {
    new = {},
    remove = {},
    reset = {},
  }
  -- listen to the creation / removal of the player's body
  -- and keep the gameState.physics.player field up to date
  local new = gameState.physics.listeners.new
  new[#new + 1] = function(gameState, object, message)
    if message == "player" then
      gameState.physics.player = object
    end
    -- track object as an entity if it has an update function
    if object.update then
      gameState.entities:add(object)
    end
  end
  local remove = gameState.physics.listeners.remove
  remove[#remove + 1] = function(gameState, object, message)
    if message == "player" then
      gameState.physics.player = nil
    end
  end
  local reset = gameState.physics.listeners.reset
  reset[#reset + 1] = function(gameState)
    gameState.physics.player = nil
  end
  -- functions that fire all listeners of the same name
  -- whenever a body is created or removed the callback
  -- should be fired
  -- similarly if the world is reset the callback should be fired
  gameState.physics.callbacks = {
    reset = function(gameState)
      -- call all listeners listening to world resets
      for _, listener in pairs(gameState.physics.listeners.reset) do
        listener(gameState)
      end
    end,
    new = function(gameState, object, message)
      -- call all listeners for newly created physics objects
      for _, listener in pairs(gameState.physics.listeners.new) do
        listener(gameState, object, message)
      end
    end,
    remove = function(gameState, object, message)
      -- call all listeners for physics objects about to be removed
      -- this should be fired before the object is actually removed
      for _, listener in pairs(gameState.physics.listeners.remove) do
        listener(gameState, object, message)
      end
    end
  }
end

function world.reload(gameState)
  -- reset world
  gameState.physics.world = wf.newWorld(0, 0, true)
  gameState.physics.callbacks.reset(gameState)
  if gameState.fsm.state == "menu" then
    -- load menu world
    local world = gameState.physics.world
    world:setGravity(0, 100)

    local player = factory.ship(gameState, 400, 0, 30)

    -- TODO move to factory
    local ground = world:newRectangleCollider(0, 550, 800, 50)
    local wall_left = world:newRectangleCollider(0, 0, 50, 600)
    local wall_right = world:newRectangleCollider(750, 0, 50, 600)
    -- make ground/walls fixed in position
    ground:setType('static')
    wall_left:setType('static')
    wall_right:setType('static')
    gameState.physics.callbacks.new(gameState, ground, "terrain")
    gameState.physics.callbacks.new(gameState, wall_left, "terrain")
    gameState.physics.callbacks.new(gameState, wall_right, "terrain")
  end
end

function world.update(gameState)
  -- update all tracked entities, going
  -- backwards so they can be removed from the list
  -- in the same pass.
  for k, entity in gameState.entities:iterateBackwards() do
    -- check if the update function returns true,
    -- if so then remove the entity
    if entity:update() then
      gameState.entities:remove(k)
    end
  end
  -- run the box2d world timestep
  gameState.physics.world:update(WORLD_TICK_TIMESTEP)
end

return world
